@@ -18,7 +18,11 @@ let push_name x = function
       { x with Genintern.genv = Environ.push_named decl x.Genintern.genv }
   | _ -> x
 
+[%%if coq = "8.19"]
 let push_gdecl (name,_,_,_) x = push_name x name
+[%%else]
+let push_gdecl (name,_,_,_,_) x = push_name x name
+[%%endif]
 
 let push_glob_ctx glob_ctx x =
   List.fold_right push_gdecl glob_ctx x
@@ -42,15 +46,34 @@ let intern_global_constr_ty { Ltac_plugin.Tacintern.genv = env } ~intern_env ?(e
 
 let intern_global_context { Ltac_plugin.Tacintern.genv = env } ~intern_env ctx =
   Constrintern.intern_context env ~bound_univs:UnivNames.empty_binders intern_env ctx
-let intern_global_context_synterp (ctx : Constrexpr.local_binder_expr list) : Glob_term.glob_decl list =
+
+[%%if coq = "8.19"]
+let intern_one h =
   let open Constrexpr in
-  let intern_one h =
-    match h with
-    | CLocalAssum(nl,Default bk,_) -> nl |> List.map (fun n -> n.CAst.v,bk,None,mkGHole)
-    | CLocalAssum(nl,Generalized(bk,_),_) -> nl |> List.map (fun n -> n.CAst.v,bk,None,mkGHole)
-    | CLocalDef (n,_,None) -> [n.CAst.v,Glob_term.Explicit,None,mkGHole]
-    | CLocalDef (n,_,Some _) -> [n.CAst.v,Glob_term.Explicit,Some mkGHole,mkGHole]
-    | CLocalPattern _ -> nYI "irrefutable pattern in synterp" in
+  match h with
+  | CLocalAssum(nl,Default bk,_) -> nl |> List.map (fun n -> n.CAst.v,bk,None,mkGHole)
+  | CLocalAssum(nl,Generalized(bk,_),_) -> nl |> List.map (fun n -> n.CAst.v,bk,None,mkGHole)
+  | CLocalDef (n,_,None) -> [n.CAst.v,Glob_term.Explicit,None,mkGHole]
+  | CLocalDef (n,_,Some _) -> [n.CAst.v,Glob_term.Explicit,Some mkGHole,mkGHole]
+  | CLocalPattern _ -> nYI "irrefutable pattern in synterp"
+[%%else]
+let intern_one h =
+  let open Constrexpr in
+  match h with
+  | CLocalAssum(nl,_,Default bk,_) -> nl |> List.map (fun n -> n.CAst.v,None,bk,None,mkGHole)
+  | CLocalAssum(nl,_,Generalized(bk,_),_) -> nl |> List.map (fun n -> n.CAst.v,None,bk,None,mkGHole)
+  | CLocalDef (n,_,_,None) -> [n.CAst.v,None,Glob_term.Explicit,None,mkGHole]
+  | CLocalDef (n,_,_,Some _) -> [n.CAst.v,None,Glob_term.Explicit,Some mkGHole,mkGHole]
+  | CLocalPattern _ -> nYI "irrefutable pattern in synterp"
+[%%endif]
+
+[%%if coq = "8.19"]
+let drop_relevance x = x
+[%%else]
+let drop_relevance (a,_,c,d,e) = (a,c,d,e)
+[%%endif]
+  
+let intern_global_context_synterp (ctx : Constrexpr.local_binder_expr list) : Glob_term.glob_decl list =
   CList.concat_map intern_one ctx |> List.rev
 
 module Cmd = struct
@@ -208,9 +231,10 @@ let univpoly_of ~poly ~cumulative =
   | true, false -> Poly
   | false, _ -> Mono
 
+[%%if coq = "8.19" || coq = "8.20"]
   let of_coq_inductive_definition id =
     let open Vernacentries.Preprocessed_Mind_decl in
-    let { flags; typing_flags; private_ind; uniform; inductives } = id in
+    let { flags;  typing_flags; private_ind; uniform; inductives } = id in
     if List.length inductives != 1 then nYI "mutual inductives";
     let inductive = List.hd inductives in
     let (((name),(parameters,non_uniform_parameters),arity,constructors),notations) = inductive in
@@ -232,6 +256,34 @@ let univpoly_of ~poly ~cumulative =
       constructors;
       univpoly = univpoly_of ~poly ~cumulative
     }
+[%%else]
+let of_coq_inductive_definition id =
+  let open Vernacentries.Preprocessed_Mind_decl in
+  let { flags; udecl; typing_flags; private_ind; uniform; inductives } = id in
+  if List.length inductives != 1 then nYI "mutual inductives";
+  let inductive = List.hd inductives in
+  let (((name),(parameters,non_uniform_parameters),arity,constructors),notations) = inductive in
+  if notations != [] then CErrors.user_err Pp.(str "notations not supported");
+  let name = [Names.Id.to_string name.CAst.v] in
+  let constructors =
+        List.map (function (Vernacexpr.(_,NoCoercion,NoInstance),c) -> c
+          | _ -> CErrors.user_err Pp.(str "coercion and instance flags not supported"))
+          constructors in
+  let { ComInductive.template; cumulative; poly; finite } = flags in
+  if template <> None then nYI "raw template polymorphic inductives";
+  if udecl <> None then nYI "raw universe polymorphic inductives with universe declaration";
+  {
+    finiteness = finite;
+    name;
+    parameters;
+    non_uniform_parameters;
+    arity;
+    constructors;
+    univpoly = univpoly_of ~poly ~cumulative
+  }
+[%%endif]
+
+[%%if coq = "8.19" || coq = "8.20"]
   let of_coq_record_definition id =
     let open Vernacentries.Preprocessed_Mind_decl in
     let { flags; primitive_proj; kind; records; } : record = id in
@@ -254,10 +306,42 @@ let univpoly_of ~poly ~cumulative =
       constructor = Some idbuild;
       fields = cfs;
       univpoly = univpoly_of ~poly ~cumulative
-    } 
-
+    }   
+[%%else]
+let of_coq_record_definition id =
+  let open Vernacentries.Preprocessed_Mind_decl in
+  let { flags; udecl; primitive_proj; kind; records; } : record = id in
+  if List.length records != 1 then nYI "mutual inductives";
+  let open Record.Ast in
+  let { name; is_coercion; binders : Constrexpr.local_binder_expr list; cfs; idbuild; sort; default_inhabitant_id : Names.Id.t option; } = List.hd records in
+  if is_coercion = Vernacexpr.AddCoercion then CErrors.user_err Pp.(str "coercion flag not supported");
+  let name = [Names.Id.to_string name.CAst.v] in
+  let sort = sort |> Option.map (fun sort ->
+    match sort.CAst.v with
+    | Constrexpr.CSort s -> s
+    | _ -> CErrors.user_err ?loc:sort.CAst.loc Pp.(str "only explicits sorts are supported")) in
+  let { ComInductive.template; cumulative; poly; finite } = flags in
+  if template <> None then nYI "raw template polymorphic inductives";
+  if udecl <> None then nYI "raw universe polymorphic inductives with universe declaration";
+  {
+    name;
+    parameters = binders;
+    sort;
+    constructor = Some idbuild;
+    fields = cfs;
+    univpoly = univpoly_of ~poly ~cumulative
+  } 
+[%%endif]
 let intern_record_decl glob_sign (it : raw_record_decl) = glob_sign, it
 
+[%%if coq = "8.19"]
+let mkCLocalAssum x y z = Constrexpr.CLocalAssum(x,y,z)
+let dest_entry (_,_,_,x) = x
+[%%else]
+let mkCLocalAssum x y z = Constrexpr.CLocalAssum(x,None,y,z)
+let dest_entry (_,_,_,_,x) = x
+[%%endif]
+
 let raw_record_decl_to_glob_synterp ({ name; sort; parameters; constructor; fields; univpoly } : raw_record_decl_elpi) : glob_record_decl_elpi =
   let name, space = sep_last_qualid name in
   let params = intern_global_context_synterp parameters in
@@ -270,20 +354,26 @@ let raw_record_decl_to_glob_synterp ({ name; sort; parameters; constructor; fiel
         if pr <> None then Coq_elpi_utils.nYI "priority in record fields";
         let atts = { Coq_elpi_HOAS.is_canonical = canon; is_coercion = if inst = Vernacexpr.AddCoercion then Reversible else Off; name } in
         let x = if bl = [] then x else Constrexpr_ops.mkCProdN bl x in
-        let entry = intern_global_context_synterp [Constrexpr.CLocalAssum ([fn],Constrexpr.Default Glob_term.Explicit,x)] in
+        let entry = intern_global_context_synterp [mkCLocalAssum [fn] (Constrexpr.Default Glob_term.Explicit) x] in
         let x = match entry with
-          | [_,_,_,x] -> x
+          | [x] -> dest_entry x
           | _ -> assert false in
         (x, atts) :: acc
     | Vernacexpr.DefExpr _, _ -> Coq_elpi_utils.nYI "DefExpr")
         [] fields in
   { name = (space, Names.Id.of_string name); arity; params; constructorname = constructor; fields = List.rev fields; univpoly }
 
+[%%if coq = "8.19"]
+let expr_Type_sort = Glob_term.UAnonymous {rigid=UState.univ_rigid}
+[%%else]
+let expr_Type_sort = Constrexpr_ops.expr_Type_sort
+[%%endif]
+
 let raw_record_decl_to_glob glob_sign ({ name; sort; parameters; constructor; fields; univpoly } : raw_record_decl_elpi) : glob_record_decl_elpi =
   let name, space = sep_last_qualid name in
   let sort = match sort with
     | Some x -> Constrexpr.CSort x
-    | None -> Constrexpr.(CSort (Glob_term.UAnonymous {rigid=UnivRigid})) in
+    | None -> Constrexpr.(CSort expr_Type_sort) in
   let intern_env, params = intern_global_context glob_sign ~intern_env:Constrintern.empty_internalization_env parameters in
   let glob_sign_params = push_glob_ctx params glob_sign in
   let params = List.rev params in
@@ -295,9 +385,9 @@ let raw_record_decl_to_glob glob_sign ({ name; sort; parameters; constructor; fi
         if pr <> None then Coq_elpi_utils.nYI "priority in record fields";
         let atts = { Coq_elpi_HOAS.is_canonical = canon; is_coercion = if inst = Vernacexpr.AddCoercion then Reversible else Off; name } in
         let x = if bl = [] then x else Constrexpr_ops.mkCProdN bl x in
-        let intern_env, entry = intern_global_context ~intern_env gs [Constrexpr.CLocalAssum ([fn],Constrexpr.Default Glob_term.Explicit,x)] in
+        let intern_env, entry = intern_global_context ~intern_env gs [mkCLocalAssum [fn] (Constrexpr.Default Glob_term.Explicit) x] in
         let x = match entry with
-          | [_,_,_,x] -> x
+          | [x] -> dest_entry x
           | _ -> assert false in
         let gs = push_glob_ctx entry gs in
         gs, intern_env, (x, atts) :: acc
@@ -325,7 +415,7 @@ let raw_indt_decl_to_glob glob_sign ({ finiteness; name; parameters; non_uniform
   let name = Names.Id.of_string name in
   let indexes = match arity with
     | Some x -> x
-    | None -> CAst.make Constrexpr.(CSort (Glob_term.UAnonymous {rigid=UnivRigid})) in
+    | None -> CAst.make Constrexpr.(CSort expr_Type_sort) in
   let intern_env, params = intern_global_context glob_sign ~intern_env:Constrintern.empty_internalization_env parameters in
   let nuparams_given, nuparams =
     match non_uniform_parameters with
@@ -363,6 +453,12 @@ let raw_decl_name_to_glob name =
   let name, space = sep_last_qualid name in
   (space, Names.Id.of_string name)
 
+[%%if coq = "8.19"]
+let interp_red_expr = Ltac_plugin.Tacinterp.interp_redexp
+[%%else]
+let interp_red_expr = Redexpr.interp_redexp_no_ltac
+[%%endif]
+
 let raw_constant_decl_to_constr ~depth coq_ctx state { name; typ = (bl,typ); body; red; udecl; atts } =
   let env = coq_ctx.env in
   let poly =
@@ -383,7 +479,7 @@ let raw_constant_decl_to_constr ~depth coq_ctx state { name; typ = (bl,typ); bod
   let sigma = get_sigma state in
   match body, typ with
   | Some body, _ ->
-      let sigma, red = option_map_acc (Ltac_plugin.Tacinterp.interp_redexp env) sigma red in
+      let sigma, red = option_map_acc (interp_red_expr env) sigma red in
       let sigma, (body, typ), impargs =
         ComDefinition.interp_definition ~program_mode:false
           env sigma Constrintern.empty_internalization_env bl red body typ
@@ -623,6 +719,7 @@ let do_record_synterp ~depth ~name ~constructorname arity fields state =
   state, in_elpi_indtdecl_record (Name.Name qrecord_name) arity constructor fields
 
 let grecord2lp_synterp ~depth state { Cmd.name; arity; params; constructorname; fields; univpoly } =
+  let params = List.map drop_relevance params in
   let state, r = do_params_synterp ~depth params (do_record_synterp ~depth ~name ~constructorname arity fields) state in
   mk_indt_decl state univpoly r
       
@@ -673,6 +770,8 @@ let drop_unit f ~depth state =
   
 let ginductive2lp_synterp ~depth state { Cmd.finiteness; name; arity; params; nuparams; nuparams_given; constructors; univpoly } =
   let space, indt_name = name in
+  let nuparams = List.map drop_relevance nuparams in
+  let params = List.map drop_relevance params in
   let do_constructor ~depth state (name, ty) =
     let state, ty = do_params_synterp nuparams (fun state -> state, in_elpi_arity E.mkDiscard) ~depth state in
     state, in_elpi_indtdecl_constructor (Name.Name name) ty
@@ -689,7 +788,9 @@ let ginductive2lp_synterp ~depth state { Cmd.finiteness; name; arity; params; nu
 let ginductive2lp ~depth state { Cmd.finiteness; name; arity; params; nuparams; nuparams_given; constructors; univpoly } =
   let open Coq_elpi_glob_quotation in
   let space, indt_name = name in
-  let contract state x = contract_params (get_global_env state) (get_sigma state) indt_name params nuparams_given x in
+  let contract state x =
+    let params = List.map drop_relevance params in
+    contract_params (get_global_env state) (get_sigma state) indt_name params nuparams_given x in
   let do_constructor ~depth state (name, ty) =
     let state, ty = do_params nuparams (do_arity (contract state ty)) ~depth state in
     state, in_elpi_indtdecl_constructor (Name.Name name) ty
@@ -719,6 +820,7 @@ let cdeclc = E.Constants.declare_global_symbol "const-decl"
 let ucdeclc = E.Constants.declare_global_symbol "upoly-const-decl"
 
 let gdecl2lp_synterp ~depth state { Cmd.name; params; typ : _; body; udecl } =
+  let params = List.map drop_relevance params in
   let state, typ = do_params_synterp ~depth params (fun state -> state, in_elpi_arity E.mkDiscard) state in
   let body = Option.map (fun _ -> E.mkDiscard) body in
   let name = decl_name2lp name in
@@ -746,6 +848,7 @@ let cdecl2lp ~depth state { Cmd.name; params; typ; body; udecl } =
 let ctxitemc = E.Constants.declare_global_symbol "context-item"
 let ctxendc =  E.Constants.declare_global_symbol "context-end"
 
+[%%if coq = "8.19"]
 let rec do_context_glob_synterp fields ~depth state =
   match fields with
   | [] -> state, E.mkGlobal ctxendc
@@ -756,7 +859,20 @@ let rec do_context_glob_synterp fields ~depth state =
       let state, bo, _ = in_option ~depth state bo in
       let state, imp = in_elpi_imp ~depth state bk in
       state, E.mkApp ctxitemc (in_elpi_id name) [imp;ty;bo;E.mkLam fields]
+[%%else]
+let rec do_context_glob_synterp fields ~depth state =
+  match fields with
+  | [] -> state, E.mkGlobal ctxendc
+  | (name,_,bk,bo,ty) :: fields ->
+      let ty = E.mkDiscard in
+      let bo = Option.map (fun _ -> E.mkDiscard) bo in
+      let state, fields = do_context_glob_synterp fields ~depth state in
+      let state, bo, _ = in_option ~depth state bo in
+      let state, imp = in_elpi_imp ~depth state bk in
+      state, E.mkApp ctxitemc (in_elpi_id name) [imp;ty;bo;E.mkLam fields]
+[%%endif]
 
+[%%if coq = "8.19"]
 let rec do_context_glob fields ~depth state =
   match fields with
   | [] -> state, E.mkGlobal ctxendc
@@ -768,7 +884,19 @@ let rec do_context_glob fields ~depth state =
       let state, bo, _ = in_option ~depth state bo in
       let state, imp = in_elpi_imp ~depth state bk in
       state, E.mkApp ctxitemc (in_elpi_id name) [imp;ty;bo;E.mkLam fields]
-
+[%%else]
+let rec do_context_glob fields ~depth state =
+  match fields with
+  | [] -> state, E.mkGlobal ctxendc
+  | (name,_,bk,bo,ty) :: fields ->
+      let open Coq_elpi_glob_quotation in
+      let state, ty = gterm2lp ~depth state ty in
+      let state, bo = option_map_acc (gterm2lp ~depth) state bo in
+      let state, fields, () = under_ctx name ty bo (nogls (do_context_glob fields)) ~depth state in
+      let state, bo, _ = in_option ~depth state bo in
+      let state, imp = in_elpi_imp ~depth state bk in
+      state, E.mkApp ctxitemc (in_elpi_id name) [imp;ty;bo;E.mkLam fields]
+[%%endif]
 let rec do_context_constr coq_ctx csts fields ~depth state =
   let map s x = constr2lp coq_ctx csts ~depth s (EConstr.of_constr x) in
   match fields with
@@ -807,9 +935,9 @@ let to_list v =
 
 (* if we make coq elaborate an arity, we get a type back. here we try to
    recoved an arity to pass that to elpi *)
+
 let best_effort_recover_arity ~depth state glob_sign typ bl =
   let _, grouped_bl = intern_global_context glob_sign ~intern_env:Constrintern.empty_internalization_env bl in
-  
   let rec aux ~depth state typ gbl =
     match gbl with
     | (name,ik,_,_) :: gbl ->
@@ -822,7 +950,7 @@ let best_effort_recover_arity ~depth state glob_sign typ bl =
       end
     | _ -> state, in_elpi_arity typ
     in
-      aux ~depth state typ (List.rev grouped_bl)
+      aux ~depth state typ (List.map drop_relevance (List.rev grouped_bl))
 
 let in_elpi_string_arg ~depth state x = 
   state, E.mkApp strc (CD.of_string x) [], []
@@ -921,6 +1049,15 @@ let handle_template_polymorphism = function
   | Some false -> Some false
   | Some true -> err Pp.(str "#[universes(template)] is not supported")
 
+[%%if coq = "8.19" || coq = "8.20"]
+let handle_template_polymorphism flags =
+  let open Vernacentries.Preprocessed_Mind_decl in
+  { flags with template = handle_template_polymorphism flags.template }
+[%%else]
+let handle_template_polymorphism flags =
+  { flags with ComInductive.template = handle_template_polymorphism flags.ComInductive.template }
+[%%endif]
+
 let in_elpi_cmd_synterp ~depth ?calldepth state (x : Cmd.raw) =
   let open Cmd in
   match x with
@@ -946,6 +1083,43 @@ let in_elpi_cmd_synterp ~depth ?calldepth state (x : Cmd.raw) =
   | Term raw_term ->
       state, E.mkApp trmc E.mkDiscard [], []
 
+[%%if coq = "8.19" || coq = "8.20"]
+let dest_rdecl raw_rdecl =
+  let open Vernacentries.Preprocessed_Mind_decl in
+  let { flags = ({ template; poly; cumulative; udecl; finite } as flags); primitive_proj; kind; records } = raw_rdecl in
+  flags, udecl, primitive_proj, kind, records
+let interp_structure ~flags udecl kind ~primitive_proj x =
+  let open Vernacentries.Preprocessed_Mind_decl in
+  let { template; poly; cumulative; finite } = flags in
+  Record.interp_structure ~template udecl kind ~cumulative ~poly ~primitive_proj finite x
+[%%else]
+let dest_rdecl (raw_rdecl : Cmd.raw_record_decl) =
+  let open Vernacentries.Preprocessed_Mind_decl in
+  let { flags; udecl; primitive_proj; kind; records } = raw_rdecl in
+  flags, udecl, primitive_proj, kind, records
+let interp_structure ~flags udecl kind ~primitive_proj x =
+  Record.interp_structure ~flags udecl kind ~primitive_proj x
+[%%endif]
+
+[%%if coq = "8.19" || coq = "8.20"]
+let dest_idecl raw_indt =
+  let open Vernacentries.Preprocessed_Mind_decl in
+  let { flags = ({ udecl } as flags); typing_flags; uniform; private_ind; inductives } = raw_indt in
+  flags, udecl, typing_flags, uniform, private_ind, inductives
+let interp_mutual_inductive ~flags ~env ~uniform ~private_ind ?typing_flags ~udecl x =
+  let open Vernacentries.Preprocessed_Mind_decl in
+  let { template; poly; cumulative; finite } = flags in
+  ComInductive.interp_mutual_inductive ~env ~template ~cumulative ~poly ~uniform ~private_ind ?typing_flags udecl x finite
+[%%else]
+let dest_idecl raw_indt =
+  let open Vernacentries.Preprocessed_Mind_decl in
+  let { flags; udecl; typing_flags; uniform; private_ind; inductives } = raw_indt in
+  flags, udecl, typing_flags, uniform, private_ind, inductives
+let interp_mutual_inductive ~flags ~env ~uniform ~private_ind ?typing_flags ~udecl x =
+  ComInductive.interp_mutual_inductive ~env ~flags ~uniform ~private_ind ?typing_flags udecl x
+[%%endif]
+
+
 let in_elpi_cmd ~depth ?calldepth coq_ctx state ~raw (x : Cmd.top) =
   let open Cmd in
   let hyps = [] in
@@ -957,12 +1131,11 @@ let in_elpi_cmd ~depth ?calldepth coq_ctx state ~raw (x : Cmd.top) =
       let state, t = grecord2lp ~depth state glob_rdecl in
       state, t, []
   | RecordDecl (_ist,(glob_sign,raw_rdecl)) ->
-      let open Vernacentries.Preprocessed_Mind_decl in
-      let { flags = { template; poly; cumulative; udecl; finite }; primitive_proj; kind; records } = raw_rdecl in
-      let template = handle_template_polymorphism template in
+      let flags, udecl, primitive_proj, kind, records = dest_rdecl raw_rdecl in
+      let flags = handle_template_polymorphism flags in
       (* Definitional type classes cannot be interpreted using this function (why?) *)
       let kind = if kind = Vernacexpr.Class true then Vernacexpr.Class false else kind in
-      let e = Record.interp_structure ~template udecl kind ~cumulative ~poly ~primitive_proj finite records in
+      let e = interp_structure ~flags udecl kind ~primitive_proj records in
       record_entry2lp ~depth coq_ctx E.no_constraints state ~loose_udecl:(udecl = None) e
   | IndtDecl (_ist,(glob_sign,raw_indt)) when raw ->
       let raw_indt = of_coq_inductive_definition raw_indt in
@@ -971,15 +1144,12 @@ let in_elpi_cmd ~depth ?calldepth coq_ctx state ~raw (x : Cmd.top) =
       let state, t = ginductive2lp ~depth state glob_indt in
       state, t, []
   | IndtDecl (_ist,(glob_sign,raw_indt)) -> 
-      let open Vernacentries.Preprocessed_Mind_decl in
-      let { flags = { template; poly; cumulative; udecl; finite }; typing_flags; uniform; private_ind; inductives } = raw_indt in
-      let template = handle_template_polymorphism template in
+      let flags, udecl, typing_flags, uniform, private_ind, inductives = dest_idecl raw_indt in
+      let flags = handle_template_polymorphism flags in
       let e =
         match inductives with
         | [mind_w_not] ->
-            ComInductive.interp_mutual_inductive ~env:coq_ctx.env
-              ~template ~cumulative ~poly ~uniform ~private_ind ?typing_flags
-              udecl [mind_w_not] finite
+            interp_mutual_inductive ~flags ~env:coq_ctx.env ~uniform ~private_ind ?typing_flags ~udecl [mind_w_not]
         | _ -> nYI "(HOAS) mutual inductives"
       in
       inductive_entry2lp ~depth coq_ctx E.no_constraints state ~loose_udecl:(udecl = None) e
