@@ -20,8 +20,13 @@ val err : ?loc:Elpi.API.Ast.Loc.t -> Pp.t -> 'a
 exception LtacFail of int * Pp.t
 val ltac_fail_err : ?loc:Elpi.API.Ast.Loc.t -> int -> Pp.t -> 'a
 val nYI : string -> 'a
+[%%if coq = "8.19"]
 val safe_destApp : Evd.evar_map ->
   EConstr.t -> (EConstr.t,EConstr.types,EConstr.ESorts.t, EConstr.EInstance.t) Constr.kind_of_term * EConstr.t array
+[%%else]
+val safe_destApp : Evd.evar_map ->
+  EConstr.t -> (EConstr.t,EConstr.types,EConstr.ESorts.t, EConstr.EInstance.t, EConstr.ERelevance.t) Constr.kind_of_term * EConstr.t array
+[%%endif]
 val mkGHole : Glob_term.glob_constr
 val pp2string : (Format.formatter -> 'a -> unit) -> 'a -> string
 val mkApp : depth:int -> Elpi.API.RawData.term -> Elpi.API.RawData.term list -> Elpi.API.RawData.term
@@ -45,8 +50,19 @@ val fold_elpi_term :
 val uint63 : Uint63.t Elpi.API.Conversion.t
 val float64 : Float64.t Elpi.API.Conversion.t
 val projection : Names.Projection.t Elpi.API.Conversion.t
+[%%if coq = "8.19"]
+val pstring : string Elpi.API.Conversion.t
+val pstring_of_string : string -> string option
+val string_of_pstring : string -> string
+[%%else]
+val pstring : Pstring.t Elpi.API.Conversion.t
+val pstring_of_string : string -> Pstring.t option
+val string_of_pstring : Pstring.t -> string
+[%%endif]
 
 val debug : CDebug.t
+val elpitime : CDebug.t
+val elpitime_flag : CDebug.flag
 
 type clause_scope = Local | Regular | Global | SuperGlobal
 val pp_scope : Format.formatter -> clause_scope -> unit
@@ -70,3 +86,5 @@ val coq_version_parser : string -> int * int * int
 
 val mp2path: Names.ModPath.t -> string list
 val gr2path: Names.GlobRef.t -> string list
+
+val eta_contract : Environ.env -> Evd.evar_map -> EConstr.t -> EConstr.t
