@@ -33,7 +33,9 @@ let atts2impl loc phase ~depth state atts q =
   and convert_att att = convert_att_r att.CAst.v
   and convert_atts l = List.map convert_att l
   and convert_att_value = function
-    Attributes.FlagIdent s | Attributes.FlagString s -> AttributeString s
+    | Attributes.FlagIdent s [@if coq = "8.19" || coq = "8.20"] -> AttributeString s
+    | Attributes.FlagQualid q [@if coq <> "8.19" && coq <> "8.20"] -> AttributeString (Libnames.string_of_qualid q)
+    | Attributes.FlagString s -> AttributeString s
   in
   let phase = match phase with Summary.Stage.Interp -> "interp" | Summary.Stage.Synterp -> "synterp" in
   let atts =
@@ -176,7 +178,8 @@ let compile src =
     compile_code src
 
 let get_and_compile ?even_if_empty name : (EC.program * bool) option =
-  P.code ?even_if_empty name |> Option.map (fun src ->
+  let t = Unix.gettimeofday () in
+  let res = P.code ?even_if_empty name |> Option.map (fun src ->
     let prog = compile src in
     let new_hash = Code.hash src in
     let old_hash = try SLMap.find name !programs_tip with Not_found -> 0 in
@@ -187,14 +190,39 @@ let get_and_compile ?even_if_empty name : (EC.program * bool) option =
       | Command { raw_args } -> raw_args
       | Program { raw_args } -> raw_args
       | Tactic -> true in
-    (prog, raw_args))
+      (prog, raw_args)) in
+    Coq_elpi_utils.elpitime (fun _ -> Pp.(str(Printf.sprintf "Elpi: get_and_compile %1.4f" (Unix.gettimeofday () -. t))));
+    res
+
+[%%if coq = "8.19" || coq = "8.20"]
+let feedback_error loc ei = Feedback.(feedback (Message(Error,loc,CErrors.iprint ei)))
+[%%else]
+let feedback_error loc ei = Feedback.(feedback (Message(Error,loc,[],CErrors.iprint ei)))
+[%%endif]
 
 let run_static_check query =
   let checker = 
     let elpi = P.ensure_initialized () in
     P.assemble_units ~elpi (P.checker()) in
   (* We turn a failure into a proper error in etc/coq-elpi_typechecker.elpi *)
-  ignore (EC.static_check ~checker query)
+  let _ : bool =
+    try EC.static_check ~checker query
+    with e ->
+      let ei = Exninfo.capture e in
+      let loc = Loc.get_loc (snd ei) in
+      (* The loc is not "Loc.mergeable" with the loc of the Elpi Typecheck
+         command, hence we generate an error feedback on the right loc *)
+      feedback_error loc ei;
+      Exninfo.iraise ei
+    in
+  ()
+
+
+
+let trace_filename_gen (add_counter: string) =
+  "/tmp/traced.tmp" ^ (add_counter) ^ ".json"
+
+let trace_filename = trace_filename_gen ""
 
 let run ~static_check program query =
   let t1 = Unix.gettimeofday () in
@@ -207,17 +235,27 @@ let run ~static_check program query =
   if static_check then run_static_check query;
   let t3 = Unix.gettimeofday () in
   let leftovers = API.Setup.trace !trace_options in
+  if (!trace_options <> [] && Sys.file_exists trace_filename) then 
+    Sys.command (Printf.sprintf "mv %s %s" trace_filename (trace_filename_gen (Printf.sprintf "_%.0f" @@ Unix.gettimeofday ()))) |> ignore;
   if leftovers <> [] then
     CErrors.user_err Pp.(str"Unknown trace options: " ++ prlist_with_sep spc str leftovers);
   let exe = EC.optimize query in
   let t4 = Unix.gettimeofday () in
-  let rc = API.Execute.once ~max_steps:!max_steps exe in
-  let t5 = Unix.gettimeofday () in
-  Coq_elpi_utils.debug Pp.(fun () ->
-      str @@ Printf.sprintf
-        "Elpi: query-compilation:%1.4f static-check:%1.4f optimization:%1.4f runtime:%1.4f\n"
-        (t2 -. t1) (t3 -. t2) (t4 -. t3) (t5 -. t4));
-  rc
+  let print_debug with_error = 
+    let t5 = Unix.gettimeofday () in
+    let txt = Printf.sprintf "Elpi: query-compilation:%1.4f static-check:%1.4f optimization:%1.4f runtime:%1.4f %s\n"
+        (t2 -. t1) (t3 -. t2) (t4 -. t3) (t5 -. t4) (if with_error then "(with error)" else "(with success)") in
+    Coq_elpi_utils.debug Pp.(fun () -> str txt);
+    Coq_elpi_utils.elpitime Pp.(fun () -> str txt)
+  in
+  try
+    let rc = API.Execute.once ~max_steps:!max_steps exe in
+    print_debug false;
+    rc
+  with e ->
+    let e = Exninfo.capture e in
+    print_debug true;
+    Exninfo.iraise e
 ;;
 
 let elpi_fails program_name =
@@ -387,7 +425,7 @@ let run_in_program ?(program = current_program ()) ?(st_setup=fun x -> x) (loc,
   
   let trace_browser _opts =
     trace_options :=
-      [ "-trace-on"; "json"; "/tmp/traced.tmp.json"
+      [ "-trace-on"; "json"; trace_filename
       ; "-trace-at"; "run"; "0"; string_of_int max_int
       ; "-trace-only"; "user"
       ];
@@ -528,6 +566,7 @@ module Synterp = struct
   let run_program loc name ~atts args =
     get_and_compile name |> Option.map (fun (program, raw_args) ->
       let loc = Coq_elpi_utils.of_coq_loc loc in
+      let initial_synterp_state = Vernacstate.Synterp.freeze () in
       let query ~depth state =
         let state, args, gls = EU.map_acc
           (Coq_elpi_arg_HOAS.in_elpi_cmd_synterp ~depth)
@@ -539,7 +578,7 @@ module Synterp = struct
         state, (loc, q), gls in
       try
         let relocate, synterplog = run_and_print ~print:false ~static_check:false name program (`Fun query) in
-        relocate "NewData", synterplog
+        initial_synterp_state, relocate "NewData", synterplog
       with Coq_elpi_builtins_synterp.SynterpAction.Error x -> err x)
 
   let run_in_program ?program query =
@@ -573,11 +612,14 @@ module Interp = struct
         |> List.map (Coq_elpi_arg_HOAS.Cmd.glob (Genintern.empty_glob_sign ~strict:true env))
         |> List.map (Coq_elpi_arg_HOAS.Cmd.interp (Ltac_plugin.Tacinterp.default_ist ()) env sigma)
       in
-      let synterplog, synterm =
+      let initial_synterp_state, synterplog, synterm=
         match syndata with
-        | None -> Coq_elpi_builtins_synterp.SynterpAction.RZipper.empty, fun ~target:_ ~depth -> Stdlib.Result.Ok ET.mkDiscard
-        | Some (relocate_term,log) -> log, relocate_term in
-      let initial_synterp_state = Vernacstate.Synterp.freeze () in
+        | None ->
+          let initial_synterp_state = Vernacstate.Synterp.freeze () in
+          initial_synterp_state,
+          Coq_elpi_builtins_synterp.SynterpAction.RZipper.empty,
+          fun ~target:_ ~depth -> Stdlib.Result.Ok ET.mkDiscard
+        | Some (initial_synterp_state,relocate_term,log) -> initial_synterp_state, log, relocate_term in
       let query ~depth state =
         let state, args, gls = EU.map_acc
           (Coq_elpi_arg_HOAS.in_elpi_cmd ~depth ~raw:raw_args Coq_elpi_HOAS.(mk_coq_context ~options:(default_options ()) state))
@@ -649,11 +691,12 @@ let run_tactic_common loc ?(static_check=false) program ~main ?(atts=[]) () =
     in
   get_and_compile program |> Option.cata (fun (cprogram, _) ->
     match run ~static_check cprogram (`Fun query) with
-    | API.Execute.Success solution -> Coq_elpi_HOAS.tclSOLUTION2EVD sigma solution
+    | API.Execute.Success solution -> Coq_elpi_HOAS.tclSOLUTION2EVD ~eta_contract_solution:false sigma solution
     | API.Execute.NoMoreSteps -> CErrors.user_err Pp.(str "elpi run out of steps")
     | API.Execute.Failure -> elpi_fails program
     | exception (Coq_elpi_utils.LtacFail (level, msg)) -> tclFAILn level msg
-    | exception e -> let e = Exninfo.capture e in (Feedback.msg_debug Pp.(str "elpi lets escape exception: " ++ CErrors.print (fst e)); Exninfo.iraise e))
+    | exception (CErrors.UserError _ as e) -> let e = Exninfo.capture e in Exninfo.iraise e
+    | exception e when CErrors.noncritical e -> let e = Exninfo.capture e in (Feedback.msg_debug Pp.(str "elpi lets escape exception: " ++ CErrors.print (fst e)); Exninfo.iraise e))
   tclIDTAC
 
 let run_tactic loc program ~atts _ist args =
